
import collections
import itertools
import gc
import os
import sys
import time
import traceback

import qubx.dataGTK
import qubx.fast.data
import qubx.global_namespace
import qubx.pyenv
import qubx.settingsGTK
from qubx.accept import *
from qubx.data_types import *
from qubx.util_types import *
from qubx.GTK import *
from qubx.settings import *
from qubx.task import *
from numpy import *
import random
import cProfile
import __main__


IoverVar_Properties = [Property('', None, 'Measures short samples of data, plots Mean v. Std., to help discover total channel count.'),
                       Property('sample_dur', 10.0, 'Duration (ms) of each sample:', accept=acceptFloatGreaterThan(0.0), format='%.3g'),
                       Property('output_name', 'Mean-Std', 'Output List name:', accept=str, format=str)]

def IoverVar(wait=True, receiver=lambda: None, **kw):
    if not kw:
        response = qubx.settingsGTK.prompt_propertied('qubx_i_over_var', IoverVar_Properties,
                                                      'Plot Mean v. Std', qubx.global_namespace.QubX)
        if response is None:
            return
        else:
            return IoverVar(wait, receiver, **response)
    if wait:
        return qubx.pyenv.call_async_wait(lambda rcv: IoverVar(wait=False, receiver=rcv, **kw))
    
    # fill in any missing keywords:
    settings = dict(SettingsMgr['qubx_i_over_var'].as_pairs())
    for prop in IoverVar_Properties:
        if prop.name and not (prop.name in kw):
            try:
                kw[prop.name] = settings[prop.name]
            except KeyError:
                kw[prop.name] = prop.default
    
    qubx.pyenv.env.OnScriptable('qubx_testing.i_over_var.IoverVar(%s)' % ', '.join('%s=%s' % (k, repr(v)) for k, v in kw.iteritems()))
    task = IoverVar_Task(receiver, **kw)
    task.OnException += task_exception_to_console
    task.start()

qubx.dataGTK.Tools.register('other', 'Plot Mean v. Std...', action=lambda menuitem: IoverVar(wait=False))



def task_exception_to_console(task, typ, val, tb):
    traceback.print_exception(typ, val, tb)

class IoverVar_Task(Task):
    def __init__(self, receiver, **kw):
        Task.__init__(self, 'IoverVar')
        self.__receiver = receiver
        self.__kw = kw
        QubX = qubx.global_namespace.QubX
        self.dataview = QubX.Data.view
        self.iSignal = QubX.DataSource.signal
        segs = QubX.DataSource.get_segmentation()
        self.chunks = chunks = []
        for seg in segs:
            for chunk in seg.chunks:
                if chunk.included:
                    chunk_w_data = chunk.get_samples()
                    chunk_w_data.src_seg = seg.index
                    chunk_w_data.off_in_seg = chunk.f - seg.offset
                    chunks.append(chunk_w_data)
        self.table = QubX.Data.file.lists.show_list(kw['output_name'])
        self.table.clear()
    def run(self):
        try:
            qubx.task.Tasks.add_task(self)
            self.run_i_over_var(**self.__kw)
        finally:
            gobject.idle_add(self.__receiver)
            qubx.task.Tasks.remove_task(self)
# take short samples of (mean, std) into a list for plotting, discovering channel count
#   resample by a delta that yields desired number of levels
#   list center-x coords of each level
#   pick K at random from each level
#   measure +/-dt about each pick, find (mean,std)
# param: sample_delta, sample_dur, samples_per_level (K)
    def run_i_over_var(self, sample_dur, output_name):
        chunks = self.chunks
        if not chunks:
            return
        samp_ms = chunks[0].sampling*1e3
        half_dur_samples = int(round(sample_dur / samp_ms) / 2) or 1
        data_min = 1e10
        data_max = 1e-10
        for chunk in chunks:
            data_min = min(data_min, min(chunk.samples))
            data_max = max(data_max, max(chunk.samples))
        resample_delta = (data_max - data_min) / 50
        samples_per_level = 8
        idealizer = qubx.fast.data.IdlStim([], True, resample_delta, 0.0, samp_ms)
        tot_n = sum(chunk.n for chunk in chunks)
        read_n = 0
        # analyze each included chunk:
        for chunk in chunks:
            idealizer.add(chunk.samples)
            read_n += chunk.n
            self.progress = read_n * 33.0 / tot_n
        idealizer.done_add()
        amps = idealizer.get_amps()
        idlchunks = [idealizer.get_next_dwells(chunk.f, chunk.n) for chunk in chunks]
        by_level = collections.defaultdict(list)
        for i in xrange(len(idlchunks)):
            chunk = chunks[i]
            for f, l, c in itertools.izip(*idlchunks[i]):
                # mid = int(round((f + l)/2.0))
                mid = random.randrange(f, l+1)
                sample_f = max(chunk.f, mid-half_dur_samples)
                sample_l = min(chunk.l, sample_f + 2*half_dur_samples)
                data = chunk.samples[sample_f-chunk.f:sample_l-chunk.f+1]
                by_level[c].append((sample_f, sample_l, data.mean(), data.std()**2))
            read_n += chunks[i].n
            self.progress = read_n * 33.0 / tot_n
            del chunks[i].samples
        if samples_per_level:
            for c in by_level:
                by_level[c] = pick_random(by_level[c], samples_per_level)
        samples_flmv = []
        for c in by_level:
            samples_flmv.extend(by_level[c])
        qubx.task.idle_wait(self.__show_list, samples_flmv)
    def __show_list(self, samples_flmv):
    #    __main__.func_to_profile = lambda: self.__do_show_list(samples_flmv)
    #    cProfile.run("func_to_profile()", "/tmp/p")
    #def __do_show_list(self, samples_flmv):
        # part of why it's slow is repeatedly invalidating panels' models, so they queue their robot and wait...?
        table = self.table
        signal_name = self.dataview.signals[self.iSignal+1, 'Name']
        mean_name = '%s Mean' % signal_name
        variance_name = '%s Std' % signal_name
        for f, l, m, v in samples_flmv:
            table.append_selection(f, l, **{mean_name: m, variance_name: sqrt(v)})
        QubX = qubx.global_namespace.QubX
        QubX.Figures.Charts.add_two_plot('List', mean_name, variance_name, False, False)
        QubX.show_charts()

def pick_random(lst, k=1, destructive=True):
    """destructive: destroys contents of lst to avoid making a copy"""
    items = lst if destructive else lst[:]
    n = len(items)
    remain = k
    results = []
    while remain and n:
        pos = random.randrange(n)
        results.append(items[pos])
        items[pos] = items[n-1]
        n -= 1
        remain -= 1
    if k == 1:
        return results[0]
    return results

